

package require obstcl

namespace eval mstclobj {}

::obstcl::class ::mstclobj::mapLexer {
    variable stream  ""
    variable atend   0
    variable lineno  0
    variable curline ""
    variable curidx  0

    constructor {} {
        set stream  ""
	set atend   0
        set lineno  0
        set curline ""
        set curidx  0
    }

    method getLineNum {} {
	return $lineno
    }

    method initFromFile {mapFile} {
	set fd [open $mapFile]
	set mapcode [read $fd]
	close $fd
	$this initFromStr $mapcode
    }

    method initFromStr {mapcode} {
	set stream  ""
	set atend   0
        set lineno  0
        set curline ""
        set curidx  0

	set stream [split $mapcode \n]

	return [llength $stream]
    }

    method getRawStream {} {
	return $stream
    }

    method eof {} {
	return $atend
    }

    method nextToken {} {
	set token ""
	while {! [string length $token] && $lineno <= [llength $stream]} {
	    if {[string length $curline] == 0 || 
		[string range  $curline 0 0] == "#"} {
		set curline [string trim [lindex $stream $lineno]]
		incr lineno
	    } else {
		# handle project args which are quoted lines
		set c [string range $curline 0 0]
		if {"$c" == "\"" || "$c" == "'"} {
	            set i [string first $c [string range $curline 1 end]]
		    if {$i == -1} {
		        set token [string range $curline 1 end]
		        set curline ""
		    } else {
		        set token [string range $curline 1 $i]
		        incr i 2
		        set curline [string trim [string range $curline $i end]]
		    }
		} else {
		    if {[regexp -indices " |\t" $curline match]} {
                        set i [lindex $match 0]
                        incr i -1
                    } else {
		        set i [string length $curline]
		    }
	            set token [string range $curline 0 $i]
		    incr i
	            set curline [string trim [string range $curline $i end]]
		}
	    }
	}
	if {! [string length $token]} {
           set atend 1
	   set token end
	}
	return $token
    }

    method nextArg {} {
	set nextarg [$this nextToken]
	if {[$this eof]} {
	    set nextarg ""
        }
	return $nextarg
    }

    method nextArgToEOL {} {
	set c [string range $curline 0 0]
	if {"$c" == "\"" || "$c" == "'"} {
	    set c [string range $curline end end]
	    if {"$c" == "\"" || "$c" == "'"} {
                set result  [string range $curline 1 end-1]
	    } else {
		set result  [string range $curline 1 end]
	    }
	} else {
	    set result $curline
	}
	set curline ""
	return $result
    }

    method checkType {type val {list ""}} {
	if {[string length [string trim $val]] == 0} {
	    return 1
	}
	set val  [string tolower $val]
	set list [string tolower $list]
        switch $type {
	    opt  -
	    file -
	    dirpath -
	    text -
	    textints -
	    exprstring -
	    string    {return 1}
	    intstring {return 1}
	    double    {return [string is double  $val]}
	    int       {return [string is integer $val]}
	    bool   {switch -- [string tolower $val] {
			on - 
			off     {return 1} 
			default {return 0} 
		      }
		   }
	    truth  {switch -- [string tolower $val] {
			true - 
			false   {return 1} 
			default {return 0} 
		      }
		   }
	    list   {return [expr {1 + [lsearch $list $val]}] }
	    multi  {
		foreach subtype $list {
		    set stype [lindex $subtype 0]
		    set sargs [lindex $subtype 1]
		    if {[$this checkType $stype $val $sargs]} {
			return 1
		    }
		}
		return 0
	    }
	}
	return 0
    }

}

::obstcl::class ::mstclobj::genericMapTclObj {

    variable      myType        {}
    arrayvariable attrVals      {}
    arrayvariable attrTypes     {}
    arrayvariable attrNames     {}
    variable      subObjs       {}
    variable      validSubObjs  {}
    variable      tagAttr       {}
    variable      endTag        {}
    variable      parentObj     {}

    destructor {
	if {[string length $parentObj] && \
	    [string length [info commands $parentObj]]} {
	    $parentObj deleteSubObj $this
	}
    }

    method getType {} {
        return $myType
    }
  
    method getParent {} {
	return $parentObj
    }

    method getValidSubObjs {} {
	return $validSubObjs
    }

    method getTag {} {
        if {[string length $tagAttr]} {
 	    return $attrVals($tagAttr)
	} else {
	    return ""
	}
    }

    method getEndTag {} {
	if {[string length $endTag]} {
	    return $endTag
	} else {
	    return end
	}
    }

    method getAttr {{attr_name {}}} {
	if {[string length $attr_name]} {
	    if {[catch {set result $attrVals($attr_name)}]} {
		error "[$this getType] object has no attribute $attr_name"
	    }
	    return $result
	} else {
	    set attrvals [list]
	    foreach a [lsort [array names attrVals]] {
		lappend attrvals $a $attrVals($a)
	    }
	    return $attrvals
	}
    }

    method getAttrName {{attr_name {}}} {
	if {[string length $attr_name]} {
	    if {[catch {set result $attrNames($attr_name)}]} {
		error "[$this getType] object has no attribute $attr_name"
	    }
	    return $result
	} else {
	    set attrvals [list]
	    foreach a [lsort [array names attrNames]] {
		lappend attrvals $a $attrNames($a)
	    }
	    return $attrvals
	}
    }

    method getAttrType {{attr_name {}}} {
	if {[string length $attr_name]} {
	    if {[catch {set result $attrTypes($attr_name)}]} {
		error "[$this getType] object has no attribute $attr_name"
	    }
	    return $result
	} else {
	    return [array get attrTypes]
	}
    }

    method setAttr {attr_name val} {
	if {[catch {set attrVals($attr_name) $val}]} {
	    error "[$this getType] object has no attribute $attr_name"
	}
	return $val
    }

    method getSubObjs {} {
	return $subObjs
    }

    method addSubObj {obj} {
        if {[catch {set obj_type [$obj getType]}]} {
	    error "$obj does not exists"
        }
	if {[lsearch $validSubObjs $obj_type] == -1} {
	    error "$obj is not a valid type of $this"
        }
	lappend subObjs $obj
	return $obj
    }

    method deleteSubObj {obj} {
	if {[set idx [lsearch $subObjs $obj]] == -1} {
	    error "$obj is not a sub-object of $this"
	}
	set before [expr {$idx - 1}]
	set after  [expr {$idx + 1}]
	set subObjs [concat [lrange $subObjs 0 $before] \
		  	    [lrange $subObjs $after end]]
    }

    method dump {{offset 0}} {
	set out ""
	set spc ""
	set top ""
	for {set i 0} {$i < $offset} {incr i} {
	    append spc " "
	}
        set top $spc
	append out $spc [$this getType] \n
	append spc "  "
	incr offset 2
	foreach attr [lsort [array names attrVals]]  {
	    set val [string trim [$this getAttr $attr]]
	    set attr_type [$this getAttrType $attr]
	    if {"$attr_type" == "opt"} {
		if {"$val" == "1"} {
		    append out $spc "$attr" \n
	        }
	    } else {
		set maxstrlen 0
	        if {[catch {llength $val}]} {
		    set maxstrlen [string length $val]
		} else {
		    foreach subval $val {
		        set tmplen [string length $subval]
			if {$tmplen > $maxstrlen} {
			    set maxstrlen $tmplen
			}
		    }
		}
		if {"[string range $val 0 0]" != "#" && $maxstrlen > 0} {
		    if {"$attr_type" == "string"  || \
		        "$attr_type" == "file"    || \
		        "$attr_type" == "dirpath"} {
	                append out $spc "$attr \"[lindex $val 0]\"" \n
		    } elseif {"$attr_type" == "exprstring"} {
			set c [string range $val 0 0]
			if {"$c" == "/" || "$c" =="("} {
			    append out $spc "$attr $val" \n
			} else {
	                    append out $spc "$attr \"[lindex $val 0]\"" \n
			}
		    } elseif {"$attr_type" == "intstring"} {
			if {[string is integer -strict $val]} {
			    append out $spc "$attr $val" \n
			} else {
	                    append out $spc "$attr \"[lindex $val 0]\"" \n
			}
		    } else {
	                append out $spc "$attr $val" \n
		    }
		}
	    }
	}
	set i 1
	foreach sub $subObjs {
	    append out $spc "# $i [$sub getType] [$sub getTag]" \n
	    append out [$sub dump $offset] 
	    incr i
	}
	append out $top [$this getEndTag] \n
	return $out
    }

    method dumpToFile {file} {
	set fd [open $file w]
	puts $fd [$this dump]
	close $fd
    }

    method parseStatus {status msg} {
        if {[string length $status]} {
	    if {[string length [info command $status]]} {
		$status insert end "object [$this getType] $msg\n"
	    } elseif {! [catch {tell $status}]} {
		puts $status "object [$this getType] $msg"
	    }
	}
    }

    method validateAttr {attr_name} {
	if {[lsearch [array names attrTypes] $attr_name] >= 0} {
	    if {[string compare $attrVals($attr_name) "#none"] == 0 ||
		[string length  $attrVals($attr_name)]         == 0} {
		return [list 2 "element is null"]
	    }
	    if {[llength $attrVals($attr_name)] != \
		[llength $attrTypes($attr_name)]} {
		return [list 0 "wrong number of elements\
                    [llength $attrVals($attr_name)], should be\
                    [llength $attrTypes($attr_name)]"]
	    }
            set errors ""
	    set i 0
	    set lexObj [::mstclobj::mapLexer #auto]
	    foreach type $attrTypes($attr_name) {
 	        set arg_val [lindex $attrVals($attr_name) $i]
		if {[string length $arg_val] == 0} {
		    append errors "missing arg $i type $type "
		}
		set a_type [lindex $type 0]
		set a_misc [lindex $type 1]
	        if {! [$lexObj checkType $a_type $arg_val $a_misc]} {
		    append errors "invalid arg $i type $type - \"$arg_val\" "
	        }
	        lappend attr_val $arg_val
                incr i
	    }
	    ::obstcl::delete $lexObj
	    if {[string length $errors]} {
	        return [list 0 $errors]
	    } else {
	        return [list 1 "valid"]
	    }
        } else {
	    return [list 0 "[$this getType] has no attribute $attr_name"]
	}
    }

    method parse {lexObj {status ""}} {
	$this parseStatus $status "parsing..."
	set token [string tolower [$lexObj nextToken]]
	if {"$token" == "[$this getType]"} {
	    # skip past token name if same has object type
	    set token [string tolower [$lexObj nextToken]]
	}
	while {"$token" != "end"} {
	    # if token is a valid sub object, recursively parse it
	    if {[lsearch $validSubObjs $token] >= 0} {
		set newObj [::mstclobj::${token}TclObj #auto $this]
		#lappend subObjs $newObj
		$newObj parse $lexObj $status
	    } elseif {[lsearch [array names attrTypes] $token] >= 0} {
	        # else, token is an object attribute, get args
		set val ""
		set attrVals($token) ""
		foreach type $attrTypes($token) {
		    if {"$type" == "opt"} {
			set attrVals($token) 1
		    } elseif {"$type" == "exprstring"} {
			set val [$lexObj nextArgToEOL]
			lappend attrVals($token) $val
		    } else {
			set val [$lexObj nextArg]
			set a_type [lindex $type 0]
			set a_misc [lindex $type 1]
			if {! [$lexObj checkType $a_type $val $a_misc]} {
			    error "token $token has invalid type $type argument\
			         \"$val\" while parsing object [$this getType]\
                                 at line [$lexObj getLineNum]"
			}
			lappend attrVals($token) $val
		    }
		}
	    } else {
		error "[$this getType] object found unknown token \"$token\"\
                       at line [$lexObj getLineNum]"
	    }
	    set token [string tolower [$lexObj nextToken]]
	}
	$this parseStatus $status "successfully parsed \
		[string range [$this getTag] 0 20]"
    }

}
    

::obstcl::class ::mstclobj::mapTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {args} {
        set myType map
        array set attrVals {
            extent       {#none}
            fontset      {#none} 
            imagecolor   {#none}
            imagequality {#none}
            imagetype    {#none}
            interlace    {#none}
            name         {#none}
            scale        {#none}
            shapepath    {#none}
            symbolset    {#none}
            size         {#none}
            status       {#none}
            tile         {#none}
            transparent  {#none}
            units        {#none}
        }
        array set attrTypes {
	    extent       {double double double double}
	    fontset      {file}
            imagecolor   {int int int}
            imagequality {int}
            imagetype    {{list {png jpg jpeg wbmp}}}
            interlace    {bool}
            name         {string}
            scale        {double}
            shapepath    {dirpath}
            symbolset    {file}
            size         {int int}
            status       {bool}
            tile         {string}
            transparent  {bool}
            units        {{list {feet inches kilometers meters miles dd}}}
        }
        array set attrNames {
	    extent       {"Extent" "Min X" "Min Y" "Max X" "Max Y"}
	    fontset      {"Fontset file"}
            imagecolor   {"Image color" "Red" "Green" "Blue"}
            imagequality {"Image qual-gd1.8"}
            imagetype    {"Image type-gd1.8"}
            interlace    {"Interlace"}
            name         {"Map name"}
            scale        {"Map scale"}
            shapepath    {"Shape dir"}
            symbolset    {"Symbolset file"}
            size         {"Map size" "Rows" "Cols"}
            status       {"Status"}
            tile         {"Tile"}
            transparent  {"Transparent"}
            units        {"Units value"}
        }
        set subObjs      {}
        set validSubObjs \
	    {layer legend projection querymap reference scalebar symbol web}
	set tagAttr      name

	return $this
    }

}


::obstcl::class ::mstclobj::labelTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType    label
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
            angle                 {#none}
            antialias             {#none}
            backgroundcolor       {#none}
            backgroundshadowcolor {#none}
            backgroundshadowsize  {#none}
            buffer                {#none} 
            color                 {#none}
            font                  {#none}
            force                 {#none}
            maxsize               {#none}
            minsize               {#none}
            mindistance           {#none}
            minfeaturesize        {#none}
            offset                {#none}
            outlinecolor          {#none}
            partials              {#none}
            position              {#none}
            size                  {#none}
            shadowcolor           {#none}
            shadowsize            {#none}
            type                  {#none}
            wrap                  {#none}
        }
        array set attrTypes {
            angle                 {{multi {double {list {auto}}}}}
            antialias             {opt}
            backgroundcolor       {int int int}
            backgroundshadowcolor {int int int}
            backgroundshadowsize  {int int}
            buffer                {int} 
            color                 {int int int}
            font                  {string}
            force                 {truth}
            maxsize               {int}
            minsize               {int}
            mindistance           {int}
            minfeaturesize        {{multi {int {list {auto}}}}}
            offset                {int int}
            outlinecolor          {int int int}
            partials              {truth}
            position              {{list {ul uc ur cl cc cr ll lc lr auto}}}
            size    {{multi {int {list {tiny small medium large giant}}}}}
            shadowcolor           {int int int}
            shadowsize            {int int}
            type                  {{list {bitmap truetype}}}
            wrap                  {string}
        }
        array set attrNames {
            angle                 {"Angle"}
            antialias             {"Anti-Alias"}
            backgroundcolor       {"Bkgrnd color" "Red" "Green" "Blue"}
            backgroundshadowcolor {"Bkgrnd shadow color" "Red" "Green" "Blue"}
            backgroundshadowsize  {"Bkgrnd shadow size" "X" "Y"}
            buffer                {"Buffer size"} 
            color                 {"Color" "Red" "Green" "Blue"}
            font                  {"Font name"}
            force                 {"Force"}
            maxsize               {"Max size"}
            minsize               {"Min size"}
            mindistance           {"Min label dist"}
            minfeaturesize        {"Min feature"}
            offset                {"Offset" "X" "Y"}
            outlinecolor          {"Outline color" "Red" "Green" "Blue"}
            partials              {"Partials"}
            position              {"Position"}
            size                  {"Size"}
            shadowcolor           {"Shadow color" "Red" "Green" "Blue"}
            shadowsize            {"Shadow size" "X" "Y"}
            type                  {"Type kind"}
            wrap                  {"Wrap character"}
        }
        set subObjs      {}
        set validSubObjs {}
	set tagAttr      font

	return $this
    }

}

::obstcl::class ::mstclobj::layerTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType layer
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
            classitem        {#none}
            connection       {#none}
            connectiontype   {#none}
            data             {#none}
            description      {#none}
            footer           {#none}
            group            {#none}
            header           {#none}
            labelangleitem   {#none}
            labelcache       {#none}
            labelitem        {#none}
            labelmaxscale    {#none}
            labelminscale    {#none}
            labelsizeitem    {#none}
            maxfeatures      {#none}
            maxscale         {#none}
            minscale         {#none}
            name             {#none}
            offsite          {#none}
            postlabelcache   {#none}
            queryitem        {#none}
            symbolscale      {#none}
            status           {#none}
            tileindex        {#none}
            tileitem         {#none}
            tolerance        {#none}
            toleranceunits   {#none}
            transform        {#none}
            type             {#none}
        }
        array set attrTypes {
            classitem        {string}
            connection       {string}
            connectiontype   {{list {local sde ogr}}}
            data             {file}
            description      {string}
            footer           {file}
            group            {string}
            header           {file}
            labelangleitem   {string}
            labelcache       {bool}
            labelitem        {string}
            labelmaxscale    {double}
            labelminscale    {double}
            labelsizeitem    {string}
            maxfeatures      {int}
            maxscale         {double}
            minscale         {double}
            name             {string}
            offsite          {int}
            postlabelcache   {truth}
            queryitem        {string}
            symbolscale      {double}
            status           {{list {on off default queryonly}}}
            tileindex        {string}
            tileitem         {string}
            tolerance        {double}
            toleranceunits   \
		    {{list {inches feet miles meters kilometers dd pixels}}}
            transform        {truth}
            type    {{list {point line polyline polygon annotation raster}}}
        }
        array set attrNames {
            classitem        {"Class item"}
            connection       {"Connection"}
            connectiontype   {"Connection type"}
            data             {"Data file"}
            description      {"Description"}
            footer           {"Footer file"}
            group            {"Group name"}
            header           {"Header file"}
            labelangleitem   {"Label angle item"}
            labelcache       {"Label cache"}
            labelitem        {"Label item"}
            labelmaxscale    {"Label max scale"}
            labelminscale    {"Label min scale"}
            labelsizeitem    {"Label size item"}
            maxfeatures      {"Max features"}
            maxscale         {"Max scale"}
            minscale         {"Min scale"}
            name             {"Layer name"}
            offsite          {"Offsite"}
            postlabelcache   {"Post label cache"}
            queryitem        {"Query item"}
            symbolscale      {"Symbol scale"}
            status           {"Status"}
            tileindex        {"Tile index"}
            tileitem         {"Tile item"}
            tolerance        {"Tolerance"}
            toleranceunits   {"Tolerance units"}
            transform        {"Transform"}
            type             {"Type"}
        }
        set subObjs      {}
        set validSubObjs {class feature query}
        set tagAttr      name

        return $this
    }

}

::obstcl::class ::mstclobj::classTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType class
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
 	    backgroundcolor {#none}
	    color           {#none}
	    expression      {#none}
	    maxsize         {#none}
	    minsize         {#none}
	    name            {#none}
	    outlinecolor    {#none}
	    overlaybackgroundcolor {#none}
	    overlaymaxsize  {#none}
	    overlayminsize  {#none}
	    overlayoutlinecolor {#none}
	    overlaysize     {#none}
	    overlaysymbol   {#none}
	    size            {#none}
	    symbol          {#none}
	    text            {#none}
        }
        array set attrTypes {
 	    backgroundcolor {int int int}
	    color           {int int int}
	    expression      {exprstring}
	    maxsize         {int}
	    minsize         {int}
	    name            {string}
	    outlinecolor    {int int int}
	    overlaybackgroundcolor {int int int}
	    overlaymaxsize  {int}
	    overlayminsize  {int}
	    overlayoutlinecolor {int int int}
	    overlaysize     {int}
	    overlaysymbol   {intstring}
	    size            {int}
	    symbol          {intstring}
	    text            {exprstring}
        }
        array set attrNames {
 	    backgroundcolor {"Bkgrnd color" "Red" "Green" "Blue"}
	    color           {"Color" "Red" "Green" "Blue"}
	    expression      {"Expression"}
	    maxsize         {"Max size"}
	    minsize         {"Min size"}
	    name            {"Class name"}
	    outlinecolor    {"Outline color" "Red" "Green" "Blue"}
	    overlaybackgroundcolor {"Overlay bg color" "Red" "Green" "Blue"}
	    overlaymaxsize  {"Overlay max size"}
	    overlayminsize  {"Overlay min size"}
	    overlaybackgroundcolor {"Overlay out color" "Red" "Green" "Blue"}
	    overlaysize     {"Overlay size"}
	    overlaysymbol   {"Overlay symbol"}
	    size            {"Size height"}
	    symbol          {"Symbol number"}
	    text            {"Text"}
        }
        set subObjs      {}
        set validSubObjs {label}
        set tagAttr      name

        return $this
    }
}


::obstcl::class ::mstclobj::featureTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType feature
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
	    text            {#none}
        }
        array set attrTypes {
	    text            {string}
        }
        array set attrNames {
	    text            {"Text"}
        }
        set subObjs      {}
        set validSubObjs {points}
        set tagAttr      text

        return $this
    }
}



::obstcl::class ::mstclobj::projectionTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
	set myType projection
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    pts {}
	}
	array set attrTypes {
	    pts {text}
	}
	array set attrNames {
	    pts {"Projection attributes"}
	}
    }

    method dump {{offset 0}} {
        # if empty or "#none" then don't dump
	if { [llength $attrVals(pts)] == 0 || \
	    "[lindex $attrVals(pts) 0]" == "#none"} {
	    return
	}
        set out ""
        set spc ""
        for {set i 0} {$i < $offset} {incr i} {
            append spc " "
        }
	set top $spc
	set i 0 
	append out $spc [$this getType] \n
	append spc "  "
	foreach p $attrVals(pts) {
	    if {"[string tolower $p]" == "geographic"} {
	        append out $spc $p \n
	    } else {
	        append out $spc "\"$p\"" \n
	    }
	    incr i
	}
	append out $top end \n
    }

    method validateAttr {attr_name} {
	if {[lsearch [array names attrTypes] $attr_name] >= 0} {
	    if {[string compare $attrVals($attr_name) "#none"] == 0 ||
		[string length  $attrVals($attr_name)]         == 0} {
		return [list 2 "element is null"]
            } else {
                return [list 1 "valid"]
            }
        } else {
            return [list 0 "[$this getType] has no attribute $attr_name"]
        }

    }

    method parse {lexObj {status ""}} {
	$this parseStatus $status "parsing..."
        set token [string tolower [$lexObj nextToken]]
	set attrVals(pts) ""
        while {"[string tolower $token]" != "end"} {
	    if {! [$lexObj checkType text $token]} {
                error "token $token has invalid type $type argument\
                                 \"$val\" while parsing object [$this getType]\
                                 at line [$lexObj getLineNum]"
	    }
	    lappend attrVals(pts) $token
	    set token [$lexObj nextToken]
        }
	$this parseStatus $status "successfully parsed"
	
    }
}

::obstcl::class ::mstclobj::pointsTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
	set myType points
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    pts {}
	}
	array set attrTypes {
	    pts {textints}
	}
	array set attrNames {
	    pts {Points}
	}
    }

    method dump {{offset 0}} {
        # if empty or "#none" then don't dump
	if { [llength $attrVals(pts)] == 0 || \
	    "[lindex $attrVals(pts) 0]" == "#none"} {
	    return
	}
        set out ""
        set spc ""
        for {set i 0} {$i < $offset} {incr i} {
            append spc " "
        }
	set i 0 
	append out $spc [$this getType] 
	set spc2 "  $spc"
	foreach p $attrVals(pts) {
	    if {! ($i % 2)} {
	        append out \n $spc2 
	    }
	    incr i
	    append out $p " "
	}
	append out \n $spc end \n
    }

    method validateAttr {attr_name} {
	# only validate points in pts attribute
	if {[lsearch [array names attrTypes] $attr_name] >= 0} {
	    if {[string compare $attrVals($attr_name) "#none"] == 0 ||
		[string length  $attrVals($attr_name)]         == 0} {
		return [list 2 "element is null"]
	    }
	    set i 0
	    set errors ""
	    foreach p $attrVals($attr_name) {
		if {! [string is integer -strict $p]} {
		    append errors "invalid arg $i type int - \"$p\" " 
	 	} 
		incr i
	    }
	    if {$i == 0} {
		return [list 2 "element is null"]
	    } elseif {($i % 2) == 0} {
	        return [list 1 "valid"]
	    } else {
	        return [list 0  "point obj has odd number of points $i"]
 	    }
        } else {
	    return [list 0 "[$this getType] has no attribute $attr_name"]
	}
    }

    method parse {lexObj {status ""}} {
	$this parseStatus $status "parsing..."
        set token [$lexObj nextToken]
	set attrVals(pts) ""
	set i 0
        while {"[string tolower $token]" != "end"} {
	    if {! [$lexObj checkType int $token]} {
                error "token $token has invalid type int argument\
                              \"$token\" while parsing object [$this getType]\
                              at line [$lexObj getLineNum]"
	    }
	    lappend attrVals(pts) $token
	    incr i
	    set token [$lexObj nextArg]
        }
	if {($i % 2) == 0} {
	    $this parseStatus $status "successfully parsed"
	} else {
	    error "point obj has odd number of points $i"
 	}
	
    }
}


::obstcl::class ::mstclobj::legendTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType legend
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
	    imagecolor      {#none}
	    interlace       {#none}
	    keysize         {#none}
	    keyspacing      {#none}
	    outlinecolor    {#none}
	    position        {#none}
	    postlabelcache  {#none}
	    size            {#none}
	    spacing         {#none}
	    status          {#none}
	    transparent     {#none}
        }
        array set attrTypes {
	    imagecolor      {int int int}
	    interlace       {bool}
	    keysize         {int int}
	    keyspacing      {int int}
	    outlinecolor    {int int int}
	    position        {{list {ul uc ur ll lc lr}}}
	    postlabelcache  {truth}
	    size            {int int}
	    spacing         {int int}
	    status          {{list {on off embed}}}
	    transparent     {bool}
        }
        array set attrNames {
	    imagecolor      {"Image color" "Red" "Green" "Blue"}
	    interlace       {"Interlace"}
	    keysize         {"Keysize" "X" "Y"}
	    keyspacing      {"Keyspacing" "X" "Y"}
	    outlinecolor    {"Outline color" "Red" "Green" "Blue"}
	    position        {"Position"}
	    postlabelcache  {"Post label cache"}
	    size            {"Size" "X" "Y"}
	    spacing         {"Spacing" "X" "Y"}
	    status          {"Status"}
	    transparent     {"Transparent"}
        }
        set subObjs      {}
        set validSubObjs {label}
        set tagAttr      position

        return $this
    }
}


::obstcl::class ::mstclobj::queryTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType query
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    expression {#none}
	    template   {#none}
	}
	array set attrTypes {
	    expression {exprstring}
	    template   {file}
	}
	array set attrNames {
	    expression {"Expression"}
	    template   {"Template file/URL"}
	}
        set subObjs      {}
        set validSubObjs {join}
        set tagAttr      expression

        return $this
    }
}

::obstcl::class ::mstclobj::querymapTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType querymap
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
            color  {#none}
            size   {#none}
            status {#none}
            style  {#none}
        }
        array set attrTypes {
            color  {int int int}
            size   {int int}
            status {bool}
            style  {{list {normal hilite selected}}}
        }
        array set attrNames {
            color  {"Color" "Red" "Green" "Blue"}
            size   {"Size" "Rows" "Cols"}
            status {"Status"}
            style  {"Style"}
        }
        set subObjs      {}
        set validSubObjs {}
        set tagAttr      style

        return $this
    }
}


::obstcl::class ::mstclobj::joinTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType join
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    footer     {#none}
	    from       {#none}
	    header     {#none}
	    name       {#none}
	    table      {#none}
	    template   {#none}
	    to         {#none}
	    type       {#none}
	}
	array set attrTypes {
	    footer     {file}
	    from       {string}
	    header     {file}
	    name       {string}
	    table      {file}
	    template   {file}
	    to         {string}
	    type       {{list {multiple single}}}
	}
	array set attrNames {
	    footer     {"Footer"}
	    from       {"From item"}
	    header     {"Header"}
	    name       {"Join name"}
	    table      {"Table file"}
	    template   {"Template"}
	    to         {"To item"}
	    type       {"Type"}
	}
        set subObjs      {}
        set validSubObjs {}
        set tagAttr      name

        return $this
    }
}


::obstcl::class ::mstclobj::referenceTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType reference
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    color        {#none}
	    extent       {#none}
	    image        {#none}
	    outlinecolor {#none}
	    size         {#none}
	    status       {#none}
	}
	array set attrTypes {
	    color        {int int int}
	    extent       {double double double double}
	    image        {file}
	    outlinecolor {int int int}
	    size         {int int}
	    status       {bool}
	}
	array set attrNames {
	    color        {"Color" "Red" "Green" "Blue"}
	    extent       {"Extent" "Min X" "Min Y" "Max X" "Max Y"}
	    image        {"Image file"}
	    outlinecolor {"Outline color" "Red" "Green" "Blue"}
	    size         {"Size" "Rows" "Cols"}
	    status       {"Status"}
	}
        set subObjs      {}
        set validSubObjs {}
        set tagAttr      image

        return $this
    }
}


::obstcl::class ::mstclobj::scalebarTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType scalebar
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    backgroundcolor {#none}
	    color           {#none}
	    imagecolor      {#none}
	    interlace       {#none}
	    intervals       {#none}
	    outlinecolor    {#none}
	    position        {#none}
	    postlabelcache  {#none}
	    size            {#none}
	    status          {#none}
	    style           {#none}
	    transparent     {#none}
	    units           {#none}
	}
	array set attrTypes {
	    backgroundcolor {int int int}
	    color           {int int int}
	    imagecolor      {int int int}
	    interlace       {bool}
	    intervals       {int}
	    outlinecolor    {int int int}
	    position        {{list {ul uc ur ll lc lr}}}
	    postlabelcache  {truth}
	    size            {int int}
	    status          {{list {on off embed}}}
	    style           {{list {0 1}}}
	    transparent     {truth}
	    units           {{list {feet inches kilometers meters miles}}}
	}
	array set attrNames {
	    backgroundcolor {"Bkgrnd color" "Red" "Green" "Blue"}
	    color           {"Color" "Red" "Green" "Blue"}
	    imagecolor      {"Image color" "Red" "Green" "Blue"}
	    interlace       {"Interlace"}
	    intervals       {"Intervals"}
	    outlinecolor    {"Outline color" "Red" "Green" "Blue"}
	    position        {"Position"}
	    postlabelcache  {"Post label cache"}
	    size            {"Size" "Rows" "Cols"}
	    status          {"Status"}
	    style           {"Style"}
	    transparent     {"Transaprent"}
	    units           {"Units"}
	}
        set subObjs      {}
        set validSubObjs {label}
        set tagAttr      position

        return $this
    }
}


::obstcl::class ::mstclobj::webTclObj {
  
    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType web
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    empty        {#none}
	    error        {#none}
	    extent       {#none}
	    footer       {#none}
	    header       {#none}
	    imagepath    {#none}
	    imageurl     {#none}
	    log          {#none}
	    maxscale     {#none}
	    maxtemplate  {#none}
	    minscale     {#none}
	    mintemplate  {#none}
	    template     {#none}
	}
        array set attrTypes {
	    empty        {string}
	    error        {string}
	    extent       {int int int int}
	    footer       {string}
	    header       {string}
	    imagepath    {dirpath}
	    imageurl     {string}
	    log          {string}
	    maxscale     {double}
	    maxtemplate  {string}
	    minscale     {double}
	    mintemplate  {string}
	    template     {string}
	}
        array set attrNames {
	    empty        {"Empty URL"}
	    error        {"Error URL"}
	    extent       {"Extent" "Min X" "Min Y" "Max X" "Max Y"}
	    footer       {"Footer file"}
	    header       {"Header file"}
	    imagepath    {"Image dir"}
	    imageurl     {"Image URL dir"}
	    log          {"Log file"}
	    maxscale     {"Max scale"}
	    maxtemplate  {"Max templ file/URL"}
	    minscale     {"Min scale"}
	    mintemplate  {"Min templ file/URL"}
	    template     {"Template file/URL"}
	}
        set subObjs      {}
        set validSubObjs {}
        set tagAttr      template

        return $this
    }

}


::obstcl::class ::mstclobj::symbolsetTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {args} {
        set myType       {#symbolset}
        set subObjs      {}
        set validSubObjs {symbol}
        set tagAttr      {}
	set endTag       {#end symbolset}

        return $this
    }
}


::obstcl::class ::mstclobj::styleTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
	set myType style
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
	array set attrVals {
	    pts {}
	}
	array set attrTypes {
	    pts {textints}
	}
	array set attrNames {
	    pts {Points}
	}
    }

    method dump {{offset 0}} {
        # if empty or "#none" then don't dump
	if { [llength $attrVals(pts)] == 0 || \
	    "[lindex $attrVals(pts) 0]" == "#none"} {
	    return
	}
        set out ""
        set spc ""
        for {set i 0} {$i < $offset} {incr i} {
            append spc " "
        }
	set i 0 
	append out $spc [$this getType] 
	set spc2 "  $spc"
	foreach p $attrVals(pts) {
	    if {! ($i % 3)} {
	        append out \n $spc2
	    }
	    incr i
	    append out $p " "
	}
	append out \n $spc end \n
    }

    method validateAttr {attr_name} {
	# only validate points in pts attribute
	if {[lsearch [array names attrTypes] $attr_name] >= 0} {
	    if {[string compare $attrVals($attr_name) "#none"] == 0 ||
		[string length  $attrVals($attr_name)]         == 0} {
		return [list 2 "element is null"]
	    }
	    set i 0
	    set errors ""
	    foreach p $attrVals($attr_name) {
		if {! [string is integer -strict $p]} {
		    append errors "invalid arg $i type int - \"$p\" " 
	 	} 
		incr i
	    }
	    if {$i == 0} {
		return [list 2 "element is null"]
	    } elseif {($i % 3) == 0} {
	        return [list 1 "valid"]
	    } else {
	        return [list 0  "point obj has odd number of points $i"]
 	    }
        } else {
	    return [list 0 "[$this getType] has no attribute $attr_name"]
	}
    }

    method parse {lexObj {status ""}} {
	$this parseStatus $status "parsing..."
        set token [$lexObj nextToken]
	set attrVals(pts) ""
	set i 0
        while {"[string tolower $token]" != "end"} {
	    if {! [$lexObj checkType int $token]} {
                error "token $token has invalid type int argument\
                               \"$token\" while parsing object [$this getType]\
                               at line [$lexObj getLineNum]"
	    }
	    lappend attrVals(pts) $token
	    incr i
	    set token [$lexObj nextArg]
        }
	if {($i % 3) == 0} {
	    $this parseStatus $status "successfully parsed"
	} else {
	    error "point obj has odd number of points $i"
 	}
	$this parseStatus $status "successfully parsed"
	
    }

}


::obstcl::class ::mstclobj::symbolTclObj {

    inherit ::mstclobj::genericMapTclObj

    constructor {parent_obj} {
        set myType symbol
	set parentObj $parent_obj
	if {[string length $parent_obj]} {
	    $parent_obj addSubObj $this
	}
        array set attrVals {
	    antialias       {#none}
	    character       {#none}
	    filled          {#none}
	    font            {#none}
	    image           {#none}
	    name            {#none}
	    transparent     {#none}
	    type            {#none}
        }
        array set attrTypes {
	    antialias       {opt}
	    character       {string}
	    filled          {opt}
	    font            {string}
	    image           {file}
	    name            {string}
	    transparent     {int}
	    type            {{list {vector ellipse pixmap styled truetype}}}
        }
        array set attrNames {
	    antialias       {"Anti-alias"}
	    character       {"Character"}
	    filled          {"Filled"}
	    font            {"Font name"}
	    image           {"Image file"}
	    name            {"Symbol name"}
	    transparent     {"Transparent idx"}
	    type            {"Type"}
        }
        set subObjs      {}
        set validSubObjs {points style}
        set tagAttr      type

        return $this
    }

}



